module COREDELAYCODE_TIP (
                        input           gsr_b,  // async active low reset
                        input           sclk1,  // [1] for TX, clean 125MHz clock from PLL
                        input           pll_lock,  // lock flag from PLL
                        input           dll_lock,  // lock flag from DLL
                        input           dll_delay_diff,  // delay difference flag from DLL
                        input           rxdly_oor_lane,  // async glitchy overflow/underflow flag for RXDLY move up/down
                        output reg      dll_code_upd,  // update delay code in DLL
                        output reg [6:0]        dll_delay_code,  // DLL delay code
                        output reg      dll_valid_code,  // code valid flag
                        output reg      reset_lane,  // active low lane controller reset
                        output reg      pause_lane,  // pause FIFO controller for clock glitches after lane reset
                        output reg      load_lane,  // load DLL delay code to RXDLY
                        output reg      move_lane,  // move RXDLY up/down on rising edge
                        output          direction_lane,  // 1=up, 0=down
                        output          mv_rd_dly_lane,  // tie-high to always move RXDLY
                        output reg      cdr_start);  // active low reset in sclk1 domain for CDR controllers, wait for reset release, lane control clean, flag_ready, and DLL code valid

reg [1:0]       rst_b;  // sync-off internal reset on sclk1 domain
reg [1:0]       start_cnt;  // counter used in start-up
reg [1:0]       lock_sync;  // synchronizer of dll_lock
reg [1:0]       diff_sync;  // synchronizer of dll_delay_diff
reg [1:0]       state;  // state machine
reg [6:0]       move_cnt;  // move counter

wire reset_b = rst_b[0];

assign mv_rd_dly_lane = 1'b1;
assign direction_lane = 1'b0;  // always counting down

always @(posedge sclk1 or negedge gsr_b) begin  // TX or non-stop clock
   if (!gsr_b) rst_b <= 0;  // async-on
   else if (pll_lock) rst_b <= {1'b1, rst_b[1]};  // sync-off
end

always @(posedge sclk1 or negedge reset_b) begin  // TX clock
   if (!reset_b) begin
      cdr_start <= 1'b0;
      reset_lane <= 1'b0;  // active low reset lane controller
      pause_lane <= 1'b0;
      start_cnt <= 0;
      lock_sync <= 0;
      diff_sync <= 0;
   end
   else begin
      lock_sync <= {dll_lock, lock_sync[1]};
      diff_sync <= {dll_delay_diff, diff_sync[1]};
      
      start_cnt <= start_cnt + 1;  // always counting

      if (reset_lane && (start_cnt == 2'b11)) pause_lane <= 1'b0;  // free running from now on
      else if (!reset_lane && (start_cnt == 2'b00)) pause_lane <= 1'b1;  // pause early than reset to avoid FIFO clock glitch mess-up
      else if (!reset_lane && (start_cnt == 2'b01)) reset_lane <= 1'b1;  // lane controller reset release

      if (reset_lane && !pause_lane && dll_valid_code) cdr_start <= 1'b1;  // after the first DLL delay code is calculated, start CDR
   end
end

always @(posedge sclk1 or negedge reset_b) begin  // TX clock
   if (!reset_b) begin
      dll_code_upd <= 1'b0;  // no updating delay code
      dll_valid_code <= 1'b0;
      dll_delay_code <= 0;
      load_lane <= 1'b0;
      move_lane <= 1'b0;
      move_cnt <= 0;
      state <= 0;
   end
   else if (lock_sync[0]) case (state)  // after DLL lock
      2'b00 : if ((&start_cnt) && (diff_sync[0] || ~cdr_start)) begin  // initial state
                 state <= 2'b01;  // after delay_diff, to load state
                 dll_code_upd <= 1'b1;  // update DLL code
              end
      2'b01 : begin  // load state
                 dll_code_upd <= 1'b0;  // one cycle assertion
                 if (&start_cnt) begin
                    state <= 2'b10;  // after delay_code propagation, to move state
                    load_lane <= 1'b1;  // load DLL code to RXDLY
                    move_cnt <= 0;
                 end
              end
      2'b10 : begin  // move state
                 load_lane <= 1'b0;  // one cycle assertion
                 if (start_cnt[0]) begin
                    if (!rxdly_oor_lane) begin  // give enough time for rxdly_oor_lane to settle
                       move_lane <= 1'b1;  // keep moving down
                       move_cnt <= move_cnt + 1;  // increment counter
                    end
                    else if (&start_cnt) begin  // at underflow
                       state <= 2'b11;  // after counter max, to valid state
                       dll_valid_code <= 1'b0;  // deassert the flag for 4 cycles
                    end
                 end
                 else move_lane <= 1'b0;  // one cycle assertion
              end
      2'b11 : if (&start_cnt) begin  // valid state
                 state <= 2'b00;  // after counter max, to initial state
                 dll_valid_code <= 1'b1;  // new delay valid delay code
                 dll_delay_code <= move_cnt;
              end
   endcase
end

endmodule